use std::rc::Rc;

use super::{Command, types::Content};

// TODO: Remove this
use wasm_bindgen::prelude::*;
#[wasm_bindgen]
extern "C" {
    #[wasm_bindgen(js_namespace = console)]
    fn log(s: &str);
}

pub fn commands() -> Vec<Rc<Command>> {
    let mut available_commands: Vec<Rc<Command>> = Vec::new();

    available_commands.push(
        Rc::new(Command::new(
        "debug".to_string(),
        "Prints the args and flags".to_string(),
        vec![
            Rc::new(super::Flag::new("long".to_string(), "l".to_string(), "A long flag".to_string())),
            Rc::new(super::Flag::new("short".to_string(), "s".to_string(), "A short flag".to_string())),
        ],
        |_console, args, flags| {
            let mut output = String::from("Flags");
            for flag in flags {
                output.push_str(&format!("\n\tLong: {}\n\tShort: {}\n\tDescription: {}", flag.long, flag.short, flag.description));
            }

            output.push_str("\nArgs:");
            for arg in args {
                output.push_str(&format!("\n\t{}", arg));
            }

            Ok(Some(output))
        },
    )));

    available_commands.push(
        Rc::new(Command::new(
        "echo".to_string(),
        "Prints the args".to_string(),
        Vec::new(),
        |_console, args, _flags| {
            Ok(Some(args.join(" ")))
        },
    )));

    available_commands.push(
        Rc::new(Command::new(
        "pwd".to_string(),
        "Prints the current working directory".to_string(),
        Vec::new(),
        |console, args, _flags| {
            if args.len() > 0 {
                return Err("pwd does not take any arguments");
            }

            log(console.get_current_dir().borrow().get_name().as_str());
            log(console.get_current_dir().borrow().get_path().as_str());

            Ok(Some(console.get_current_dir().borrow().get_path()))
        },
    )));

    available_commands.push(
        Rc::new(Command::new(
            "cd".to_string(),
            "Changes the current working directory".to_string(),
            Vec::new(),
            |console, args, _flags| {
                if args.len() != 1 {
                    return Err("Only the path must be specified!");
                }

                match console.cd(args[0].clone(), None) {
                    Ok(_) => Ok(None),
                    Err(e) => Err(e),
                }
            },
    )));

    available_commands.push(
        Rc::new(Command::new(
            "ls".to_string(),
            "Lists the contents of the current directory".to_string(),
            vec![
                Rc::new(super::Flag::new("list".to_string(), "l".to_string(), "Print contents in a list".to_string())),
            ],
            |console, args, flags| {
                if args.len() > 1 {
                    return Err("ls does not take any arguments but a path");
                }

                let seperator = match flags.iter().find(|f| f.long == "list" ) {
                    Some(_) => "\n",
                    None => "\t",
                };

                let mut output = String::new();

                let base = if args.len() == 0 {
                    console.get_current_dir()
                } else {
                    match console.virtual_cd(args[0].clone(), Some(1)) {
                        Ok(dir) => dir,
                        Err(e) => return Err(e),
                    }
                };

                for dir in base.borrow().get_subdirs() {
                    output.push_str(&format!("{}{}", dir.borrow().get_name(), seperator));
                }
                for file in base.borrow().get_files() {
                    output.push_str(&format!("{}{}", file.borrow().get_name(), seperator));
                }

                Ok(Some(output))
            },
    )));

    available_commands.push(
        Rc::new(Command::new(
            "mkdir".to_string(),
            "Creates a new directory".to_string(),
            Vec::new(),
            |console, args, _flags| {
                if args.len() != 1 {
                    return Err("Only the path must be specified!");
                }

                let dir = match console.virtual_cd(args[0].clone(), Some(1)) {
                    Ok(dir) => dir,
                    Err(e) => return Err(e),
                };

                let result = match dir.borrow_mut().mkdir(args[0].clone()) {
                    Ok(_) => Ok(None),
                    Err(e) => Err(e),
                };

                result
            },
    )));

    available_commands.push(
        Rc::new(Command::new(
            "touch".to_string(),
            "Creates a new file".to_string(),
            Vec::new(),
            |console, args, _flags| {
                if args.len() != 1 {
                    return Err("Only the path must be specified!");
                }

                let dir = match console.virtual_cd(args[0].clone(), Some(1)) {
                    Ok(dir) => dir,
                    Err(e) => return Err(e),
                };

                let result = match dir.borrow_mut().touch(args[0].clone()) {
                    Ok(_) => Ok(None),
                    Err(e) => Err(e),
                };

                result
            },
    )));

    available_commands.push(
        Rc::new(Command::new(
            "cat".to_string(),
            "Prints the contents of a file".to_string(),
            Vec::new(),
            |console, args, _flags| {
                if args.len() != 1 {
                    return Err("Only the file must be specified!");
                }

                let dir = match console.virtual_cd(args[0].clone(), Some(1)) {
                    Ok(dir) => dir,
                    Err(e) => return Err(e),
                };

                let file = match dir.borrow().get_file(args[0].clone()) {
                    Some(file) => file,
                    None => return Err("File not found"),
                };

                let content = file.borrow().read();

                Ok(Some(content))
            },
    )));

    available_commands.push(
        Rc::new(Command::new(
            "rm".to_string(),
            "Removes a file or directory".to_string(),
            vec![
                Rc::new(super::Flag::new("recursive".to_string(), "r".to_string(), "Remove directories and their contents recursively".to_string())),
            ],
            |console, args, flags| {
                if args.len() != 1 {
                    return Err("Only the name must be specified!");
                }

                let dir = match console.virtual_cd(args[0].clone(), Some(1)) {
                    Ok(dir) => dir,
                    Err(e) => return Err(e),
                };

                _ = match dir.borrow_mut().remove_file(args[0].clone()) {
                    Ok(_) => return Ok(None),
                    Err(_) => "",
                };

                if flags.iter().find(|f| f.long == "recursive" ).is_some() {
                    match dir.borrow_mut().remove_dir(args[0].clone()) {
                        Ok(_) => return Ok(None),
                        Err(e) => return Err(e),
                    };
                } else {
                    return Err("File could not be found!");
                }                
            },
    )));

    // write into file
    available_commands.push(
        Rc::new(Command::new(
            "write".to_string(),
            "Writes into a file".to_string(),
            vec![
                Rc::new(super::Flag::new("append".to_string(), "a".to_string(), "Append to the file".to_string())),
            ],
            |console, args, flags| {
                if args.len() < 2 {
                    return Err("Path and content must be specified!");
                }

                log(format!("Args: {:?}", args).as_str());

                let dir = match console.virtual_cd(args[0].clone(), Some(1)) {
                    Ok(dir) => dir,
                    Err(e) => return Err(e),
                };

                let file = match dir.borrow().get_file(args[0].clone()) {
                    Some(file) => file,
                    None => return Err("File not found"),
                };

                if flags.iter().find(|f| f.long == "append" ).is_some() {
                    file.borrow_mut().append(args[1..].join(" "));
                    return Ok(None);
                } else {
                    file.borrow_mut().write(args[1..].join(" "));
                    Ok(None)
                }
                
            },
    )));

    available_commands
}